;; Generated package description from evil.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "evil" "1.15.0.0.20260728.297" "Extensible vi layer" '((emacs "24.1") (cl-lib "0.5") (goto-chg "1.6") (nadvice "0.3")) :commit "6a3e1ddd04ac504a016590940d0af2a3361b9efd" :maintainer '("Tom Dalziel" . "tom.dalziel@gmail.com") :keywords '("emulations") :url "https://github.com/emacs-evil/evil")
